# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Torrra"
copyright = "2025, stabldev"
author = "stabldev"
release = "1.2.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["myst_parser", "sphinx_copybutton", "qiskit_sphinx_theme"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "qiskit-ecosystem"
html_theme_options = {
    "source_repository": "https://github.com/stabldev/torrra",
    "source_branch": "main",
    "source_directory": "docs/",
}

templates_path = ["_templates"]
html_static_path = ["_static"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]



from typing import ClassVar

from textual import work
from textual.app import App
from textual.binding import BindingType
from textual.types import CSSPathType

from torrra._types import Indexer
from torrra.screens.welcome import WelcomeScreen
from torrra.utils.fs import get_resource_path


class TorrraApp(App[None]):
    TITLE: str | None = "torrra"
    CSS_PATH: ClassVar[CSSPathType | None] = get_resource_path("app.css")
    BINDINGS: ClassVar[list[BindingType]] = [
        ("q", "quit", "Quit"),
        ("d", "toggle_dark_mode", "Toggle dark mode"),
        ("escape", "clear_focus", "Clear focus"),
    ]
    ENABLE_COMMAND_PALETTE: ClassVar[bool] = False

    def __init__(self, provider: Indexer | None, use_cache: bool) -> None:
        super().__init__()
        self.provider: Indexer | None = provider
        self.use_cache: bool = use_cache

    @work
    async def on_mount(self) -> None:
        self.theme = (  # pyright: ignore[reportUnannotatedClassAttribute]
            "catppuccin-mocha"
        )

        if query := await self.push_screen_wait(WelcomeScreen(provider=self.provider)):
            from torrra.screens.search import SearchScreen

            await self.push_screen(
                SearchScreen(
                    indexer=self.provider, initial_query=query, use_cache=self.use_cache
                )
            )

    def action_toggle_dark_mode(self) -> None:
        self.theme = (
            "catppuccin-mocha"
            if self.theme == "catppuccin-latte"
            else "catppuccin-latte"
        )

    def action_clear_focus(self) -> None:
        self.set_focus(None)



from typing import ClassVar, override

from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Grid
from textual.screen import Screen
from textual.types import CSSPathType
from textual.widgets import Input, Static

from torrra._types import Indexer
from torrra._version import __version__
from torrra.utils.fs import get_resource_path

BANNER = """
▀█▀ █▀█ █▀▄ █▀▄ █▀▄ █▀█
 █  █ █ █▀▄ █▀▄ █▀▄ █▀█
 ▀  ▀▀▀ ▀ ▀ ▀ ▀ ▀ ▀ ▀ ▀
"""


class WelcomeScreen(Screen[str]):
    CSS_PATH: ClassVar[CSSPathType | None] = get_resource_path("screens/welcome.css")

    def __init__(self, provider: Indexer | None) -> None:
        super().__init__()
        self.provider: Indexer | None = provider

    @override
    def compose(self) -> ComposeResult:
        with Container():
            yield Static(BANNER, id="banner")
            yield Static(
                "\n".join(
                    [
                        "Find and download torrents right from here.",
                        "Powered by libtorrent and Python ❤️",
                    ]
                ),
                id="subtitle",
            )
            yield Input(placeholder="Search...", id="search")
            provider_name = f" - {self.provider.name}" if self.provider else ""
            yield Static(f"v{__version__}{provider_name}", id="version")
            with Container(id="commands_container"):
                with Grid(id="commands"):
                    yield Static("[commands]", id="title", markup=False)
                    yield Static("[esc]ape focus", markup=False)
                    yield Static("esc", classes="key")
                    yield Static("toggle [d]ark mode", markup=False)
                    yield Static("d", classes="key")
                    yield Static("[q]uit", markup=False)
                    yield Static("q", classes="key")

    @on(Input.Submitted, "#search")
    async def handle_search(self, event: Input.Submitted) -> None:
        if query := event.value:
            self.dismiss(query)



from dataclasses import asdict, dataclass
from typing import Any, Literal, TypedDict


# TORRENT TYPES
class TorrentDict(TypedDict):
    title: str
    size: float
    seeders: int
    leechers: int
    source: str
    magnet_uri: str | None


@dataclass
class Torrent:
    title: str
    size: float
    seeders: int
    leechers: int
    source: str
    magnet_uri: str | None

    @classmethod
    def from_dict(cls, d: TorrentDict) -> "Torrent":
        return cls(**d)

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)


# INDEXER TYPES
IndexerName = Literal["jackett", "prowlarr"]


@dataclass
class Indexer:
    name: IndexerName
    url: str
    api_key: str



class ConfigError(Exception):
    """Custom error for config key issues."""

    pass


class JackettConnectionError(Exception):
    """Raised when Jackett is unreachable or misconfigured."""

    pass


class ProwlarrConnectionError(Exception):
    """Raised when Prowlarr is unreachable or misconfigured."""

    pass



import click

from torrra._types import Indexer, IndexerName


def handle_indexer_command(
    name: IndexerName,
    indexer_cls_str: str,
    connection_error_cls_str: str,
    url: str | None,
    api_key: str | None,
    no_cache: bool,
):
    if url is None and api_key is None:
        from torrra.core.context import config
        from torrra.core.exceptions import ConfigError

        try:
            url = config.get(f"indexers.{name}.url")
            api_key = config.get(f"indexers.{name}.api_key")
        except ConfigError as e:
            click.secho(f"{e}\ncheck your configuration file.", fg="red", err=True)
            return
    elif url is None or api_key is None:
        click.secho(
            "both --url and --api-key must be provided together, or neither to use config.",
            fg="red",
            err=True,
        )
        return

    click.secho(f"connecting to {name} server at {url}", fg="cyan")
    import asyncio

    indexer_cls = _lazy_import(indexer_cls_str)
    connection_error_cls = _lazy_import(connection_error_cls_str)

    async def validate_indexer() -> bool:
        try:
            return await indexer_cls(url, api_key).validate()
        except connection_error_cls as e:
            click.secho(str(e), fg="red", err=True)
            return False

    if asyncio.run(validate_indexer()):
        from torrra.app import TorrraApp
        from torrra.core.context import config

        # update/store indexers configuration
        config.set(f"indexers.{name}.url", url)
        config.set(f"indexers.{name}.api_key", api_key)

        provider = Indexer(name, url, api_key)
        app = TorrraApp(provider, use_cache=not no_cache)
        app.run()


def _lazy_import(dotted_path: str):
    import importlib

    try:
        module_path, obj_name = dotted_path.rsplit(".", 1)
        module = importlib.import_module(module_path)
        return getattr(module, obj_name)
    except (ModuleNotFoundError, AttributeError) as e:
        click.secho(f"failed to load: {dotted_path}\n{e}", fg="red", err=True)
        raise



from typing import Any, cast

import httpx

from torrra._types import Torrent, TorrentDict
from torrra.core.cache import get_cache, has_cache, make_cache_key, set_cache
from torrra.core.exceptions import JackettConnectionError


class JackettIndexer:
    def __init__(self, url: str, api_key: str, timeout: int = 10):
        self.url: str = url.rstrip("/")
        self.api_key: str = api_key
        self.timeout: int = timeout

    async def search(self, query: str, use_cache: bool = True) -> list[Torrent]:
        key = make_cache_key("jackett", query)

        if use_cache and has_cache(key):
            raw_data = cast(list[TorrentDict], get_cache(key))
            return [Torrent.from_dict(d) for d in raw_data]

        endpoint = f"{self.url}/api/v2.0/indexers/all/results"
        params = {"apikey": self.api_key, "query": query}

        async with httpx.AsyncClient(timeout=self.timeout) as client:
            resp = await client.get(endpoint, params=params)
            resp.raise_for_status()
            res = resp.json().get("Results", [])
            torrents = [self._normalize_result(r) for r in res]

        if use_cache:
            set_cache(key, [t.to_dict() for t in torrents])

        return torrents

    async def validate(self) -> bool:
        url = f"{self.url}/api/v2.0/indexers/nonexistent_indexer/results"
        params = {"apikey": self.api_key}

        async with httpx.AsyncClient(timeout=self.timeout) as client:
            try:
                resp = await client.get(url, params=params)
                resp.raise_for_status()
                return True

            except httpx.RequestError:
                raise JackettConnectionError(
                    f"could not connect to jackett server\n"
                    + "please make sure jackett server is running and the url is correct"
                )

            except httpx.HTTPStatusError as e:
                status_code = e.response.status_code

                if status_code == 401:
                    raise JackettConnectionError(
                        "invalid jackett server api key\n"
                        + "double-check the api key you provided"
                    )
                elif status_code == 500 and "nonexistent_indexer" in e.response.text:
                    return True
                else:
                    raise JackettConnectionError(
                        f"jackett server returned http {status_code}\n"
                        + "unexpected response from jackett server. please verify your setup"
                    )

    def _normalize_result(self, r: dict[str, Any]) -> Torrent:
        return Torrent(
            title=r.get("Title", ""),
            size=r.get("Size", 0),
            seeders=r.get("Seeders", 0),
            leechers=r.get("Peers", 0),
            source=r.get("Tracker", "unknown"),
            magnet_uri=r.get("MagnetUri", None),
        )



import atexit
import hashlib
from typing import Any, cast

# no official stubs for diskcache
from diskcache import Cache  # pyright: ignore[reportMissingTypeStubs]
from platformdirs import user_cache_dir

from torrra.core.constants import CACHE_TTL

cache = Cache(user_cache_dir("torrra"))
atexit.register(cache.close)

# ========== CACHE FUNCTIONS ========== #


def has_cache(key: str) -> bool:
    return key in cache


def get_cache(key: str) -> Any:
    return cast(Any, cache[key])


def set_cache(key: str, value: Any, expire: int = CACHE_TTL) -> None:
    cache.set(key, value, expire=expire)  # pyright: ignore[reportUnknownMemberType]


def delete_cache(key: str) -> None:
    if key in cache:
        del cache[key]


# ========== CACHE UTILS ========== #


def make_cache_key(prefix: str, query: str) -> str:
    return f"{prefix}:{hashlib.sha256(query.encode()).hexdigest()}"
